//----------------------------------------------------------------------------//

/*
 * Copyright (c) 2014 Sony Pictures Imageworks Inc
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.  Neither the name of Sony Pictures Imageworks nor the
 * names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

//----------------------------------------------------------------------------//

/*! \file FileSequence.h
  \brief Contains the FileSequence class which can be used to turn strings
  such as "file.1-20#.f3d" into a vector of strings
*/

//----------------------------------------------------------------------------//

#ifndef _INCLUDED_Field3D_FileSequence_H_
#define _INCLUDED_Field3D_FileSequence_H_

//----------------------------------------------------------------------------//

#include <string>
#include <vector>

//----------------------------------------------------------------------------//

#include "ns.h"

FIELD3D_NAMESPACE_OPEN

//----------------------------------------------------------------------------//
// FileSequence
//----------------------------------------------------------------------------//

class FileSequence
{

public:

  // Typedefs ---

  typedef std::vector<std::string> StringVec;

  // Ctors, dtor ---

  //! Default constructor. Creates an empty sequence
  FileSequence()
  { }
  //! Construct from a sequence string. If the sequence string is an actual
  //! filename, we become a 1-length sequence. If the sequence can't be 
  //! resolved, we become a 0-length sequence. 
  //! \note Apart from checking if the sequence is itself a filename, no
  //! checks are made to see that the actual resulting files exist.
  FileSequence(const std::string &sequence);

  //! Number of files in sequence
  size_t size() const
  { return m_filenames.size(); }

  //! Returns a single filename
  const std::string& filename(const size_t idx) const
  { return m_filenames[idx]; }

  //! Returns the full list of all filenames
  const StringVec& filenames() const
  { return m_filenames; }

private:

  // Data members ------------------------------------------------------------

  //! Stores the resulting filenames
  std::vector<std::string> m_filenames;

};

//----------------------------------------------------------------------------//

FIELD3D_NAMESPACE_HEADER_CLOSE

//----------------------------------------------------------------------------//

#endif // Include guard
